"""
OpenTmiClient module
"""
# standard imports
import json
import os
# 3rd party imports
# import deprecation
# Application imports
from opentmi_client.utils import is_object_id, get_logger
# from opentmi_client.utils import requires_logged_in
from opentmi_client.utils import OpentmiException, TransportException
from opentmi_client.utils.decorators import setter_rules
from opentmi_client.transport import Transport
from opentmi_client.api.result import Result
from opentmi_client.api.build import Build


REQUEST_TIMEOUT = 30

ENV_GITHUB_ACCESS_TOKEN = "OPENTMI_GITHUB_ACCESS_TOKEN"
ENV_OPENTMI_USERNAME = "OPENTMI_USERNAME"
ENV_OPENTMI_PASSWORD = "OPENTMI_PASSWORD"


# pylint: disable-msg=too-many-arguments
def create(host='localhost', port=None, result_converter=None, testcase_converter=None):
    """
    Generic create -api for Client
    :param host:
    :param port:
    :param result_converter:
    :param testcase_converter:
    :return: OpenTmiClient
    """
    client = OpenTmiClient(host, port)
    client.set_result_converter(result_converter)
    client.set_tc_converter(testcase_converter)
    return client


class OpenTmiClient(object):
    """
    OpenTmiClient object
    """
    __version = 0
    __api = "/api/v"

    # pylint: disable-msg=too-many-arguments
    def __init__(self,
                 host='127.0.0.1',
                 port=None,
                 transport=None):
        """
        Constructor for OpenTMI client
        :param host: opentmi host address (default="localhost")
        :param port: opentmi server port (default=3000)
        :param transport: optional Transport layer. Mostly for testing purpose
        """
        self.__logger = get_logger()
        self.__transport = Transport(host, port) if not transport else transport
        # backward compatibility
        self.__result_converter = None
        self.__tc_converter = None
        self.try_login()

    def set_result_converter(self, func):
        """
        Set custom result converter
        :param func: conversion function
        :return: None
        """
        self.__result_converter = func

    def set_tc_converter(self, func):
        """
        Set custom test case converter
        :param func: conversion function
        :return: None
        """
        self.__tc_converter = func

    def login(self, username, password):
        """
        Login to OpenTMI server
        :param username: username for OpenTMI
        :param password: password for OpenTMI
        :return: OpenTmiClient
        """
        payload = {
            "email": username,
            "password": password
        }
        url = self.__resolve_url("/auth/login")
        response = self.__transport.post_json(url, payload)
        token = response.get("token")
        self.logger.info("Login success. Token: %s", token)
        self.set_token(token)
        return self

    def login_with_access_token(self, access_token, service="github"):
        """
        Login to OpenTMI server using access token
        :param access_token: access token to be used
        :param service: access token provider
        :return: OpenTmiClient
        """
        payload = {
            "access_token": access_token
        }
        url = "{}/auth/{}/token".format(self.__transport.host, service)
        self.logger.debug("Login using %s token", service)
        response = self.__transport.post_json(url, payload)
        token = response.get("token")
        self.logger.info("Login success. Token: %s", token)
        self.set_token(token)
        return self

    @property
    def is_logged_in(self):
        """
        get logged in state
        :return: boolean true if logged in.
        """
        return self.__transport.has_token()

    def set_logger(self, logger):
        """
        Set custom logger
        :param logger: logging.Logger instance
        :return: OpenTmiClient
        """
        self.__logger = logger
        return self

    @property
    def logger(self):
        """
        getter for logger
        :return: Logger
        """
        return self.__logger

    def logout(self):
        """
        Logout
        :return: OpenTmiClient
        """
        self.__transport.set_token(None)
        return self

    def set_token(self, token):
        """
        Set authentication token for transport layer
        :param token:
        :return: OpenTmiClient
        """
        self.__transport.set_token(token)
        return self

    def get_version(self):
        """
        Get Client version
        :return:
        """
        return self.__version

    # @requires_logged_in
    @setter_rules(value_type=Build)
    def post_build(self, build):
        """
        Send build
        :param build: Build object
        :return: Stored build data
        """
        payload = build.data
        url = self.__resolve_apiuri("/duts/builds")
        try:
            data = self.__transport.post_json(url, payload)
            self.logger.debug("build uploaded successfully, _id: %s", data.get("_id"))
            return data
        except TransportException as error:
            self.logger.warning("Result upload failed: %s (status: %s)", error.message, error.code)
        except OpentmiException as error:
            self.logger.warning(error)
        return None

    # @requires_logged_in
    def upload_build(self, build):
        """
        Upload build
        :param build:
        :return:
        """
        build_dict = build
        build = Build()
        build.set_data(build_dict)
        return self.post_build(build)

    # Suite
    # @requires_logged_in
    def get_suite(self, suite, options=''):
        """
        get single suite informations
        :param suite:
        :param options:
        :return:
        """
        try:
            campaign_id = self.__get_campaign_id(suite)
        except OpentmiException as error:
            self.logger.warning("exception happened while resolving suite: %s, %s",
                                suite, error)
            return None

        if campaign_id is None:
            self.logger.warning("could not resolve campaign id for suite: %s",
                                suite)
            return None

        suite = self.__get_suite(campaign_id, options)
        return suite

    # Campaign

    # @requires_logged_in
    def get_campaigns(self):
        """
        Get campaigns
        :return:
        """
        return self.__get_campaigns()

    # @requires_logged_in
    def get_campaign_names(self):
        """
        Get campaign names
        :return:
        """
        campaigns = self.__get_campaigns()
        campaign_names = []
        for campaign in campaigns:
            campaign_names.append(campaign['name'])
        return campaign_names

    # @requires_logged_in
    def get_testcases(self, filters=None):
        """
        Get testcases
        :param filters:
        :return:
        """
        return self.__get_testcases(filters)

    # @requires_logged_in
    def update_testcase(self, metadata):
        """
        update test case
        :param metadata:
        :return:
        """
        testcase = self.__lookup_testcase(metadata['tcid'])
        if testcase:
            test_id = testcase.get('_id')
            self.logger.info("Update existing TC (%s)", test_id)
            self.__update_testcase(test_id, metadata)
        else:
            self.logger.info("Create new TC")
            self.__create_testcase(metadata)
        return self

    # @requires_logged_in
    @setter_rules(value_type=Result)
    def post_result(self, result):
        """
        Post Result object
        :param result: Result or plain dictionary
        :return:
        """
        url = self.__resolve_apiuri("/results")
        payload = result.data
        try:
            files = None
            # hasLogs, logFiles = result.hasLogs()
            # if hasLogs:
            #    zipFile = self.__archiveLogs(logFiles)
            #    self.logger.debug(zipFile)
            #    files = {"file": ("logs.zip", open(zipFile), 'rb') }
            #    self.logger.debug(files)
            data = self.__transport.post_json(url, payload, files=files)
            self.logger.debug("result uploaded successfully, _id: %s", data.get("_id"))
            return data
        except TransportException as error:
            self.logger.warning("result uploaded failed: %s. status_code: %d",
                                error.message, error.code)
        except OpentmiException as error:
            self.logger.warning(error)
        return None

    # @deprecation.deprecated(deprecated_in="v0.4.0", removed_in="v0.5.0",
    #                        details="Use post_result(Result) instead")
    def upload_results(self, result):
        """
        Upload result, and test case if not stored already
        :param result: dictionary
        :return: Dictionary
        """
        tc_meta = self.__tc_converter(result.tc_metadata) if self.__tc_converter else result
        test_case = self.__lookup_testcase(tc_meta['tcid'])
        if not test_case:
            test_case = self.__create_testcase(tc_meta)
            if not test_case:
                self.logger.warning("TC creation failed")
                return None

        result_dict = self.__result_converter(result) if self.__result_converter else result
        result = Result()
        result.set_data(result_dict)
        return self.post_result(result)

    def try_login(self, raise_if_fail=False):
        """
        function to check if login is done.
        If not try to use environment variables by default
        :param raise_if_fail: Boolean, raise if login failed
        or env variables are does not exists. Default False
        :return: OpenTmiClient
        :throws: OpentmiException in case of failure
        """
        # use environment variables if available
        token = os.getenv(ENV_GITHUB_ACCESS_TOKEN)
        if token:
            self.logger.info("Using github access token from environment variable")
            return self.login_with_access_token(access_token=token, service="github")
        username = os.getenv(ENV_OPENTMI_USERNAME)
        password = os.getenv(ENV_OPENTMI_PASSWORD)
        if username and password:
            self.logger.info("Using opentmi credentials from environment variable")
            return self.login(username, password)
        if raise_if_fail:
            raise OpentmiException("login required")
        return self

    # Private members

    # @requires_logged_in
    def __get_campaign_id(self, campaign_name):
        """
        get campaign id from name
        :param campaign_name:
        :return: string
        """
        if is_object_id(campaign_name):
            return campaign_name

        for campaign in self.__get_campaigns():
            if campaign['name'] == campaign_name:
                return campaign['_id']
        return None

    def __get_testcases(self, filters=None):
        url = self.__resolve_apiuri("/testcases")
        return self.__transport.get_json(url, params=filters if filters else None)

    def __get_campaigns(self):
        url = self.__resolve_apiuri("/campaigns")
        return self.__transport.get_json(url, params={"f": "name"})

    def __get_suite(self, suite, options=''):
        url = self.__resolve_apiuri("/campaigns/" + suite + "/suite" + options)
        return self.__transport.get_json(url)

    def __lookup_testcase(self, tcid):
        url = self.__resolve_apiuri("/testcases")
        self.logger.debug("Search TC: %s", tcid)
        try:
            data = self.__transport.get_json(url, params={"tcid": tcid})
            if len(data) == 1:
                doc = data[0]
                self.logger.debug("testcase '%s' exists in DB (%s)", tcid, doc.get('_id'))
                return doc
        except TransportException as error:
            if error.code == 404:
                self.logger.warning("testcase '%s' not found form DB", tcid)
            else:
                self.logger.warning("Test case find failed: %s", error.message)
        except OpentmiException as error:
            self.logger.warning(error)

        return None

    def __update_testcase(self, test_id, metadata):
        url = self.__resolve_apiuri("/testcases/" + test_id)
        try:
            self.logger.debug("Update TC: %s", url)
            payload = metadata
            data = self.__transport.put_json(url, payload)
            self.logger.debug("testcase metadata uploaded successfully")
            return data
        except TransportException as error:
            self.logger.debug(error)
        except OpentmiException as error:
            self.logger.debug(error)

        self.logger.warning("testcase metadata upload failed")
        return None

    def __create_testcase(self, metadata):
        url = self.__resolve_apiuri("/testcases")
        try:
            self.logger.debug("Create TC: %s", url)
            payload = metadata
            data = self.__transport.post_json(url, payload)
            self.logger.debug("new testcase metadata uploaded successfully with id: %s",
                              json.dumps(data))
            return data
        except TransportException as error:
            self.logger.warning(error)
        except OpentmiException as error:
            self.logger.warning('createTestcase throw exception:')
            self.logger.warning(error)

        self.logger.warning("new testcase metadata upload failed")
        return None

    def __resolve_url(self, path):
        return self.__transport.get_url(path)

    def __resolve_apiuri(self, path):
        return self.__resolve_url(self.__api + str(self.__version) + path)



from opentmi_client import OpenTmiClient, Build
from opentmi_client.api.build import Ci, Target, Hardware


client = OpenTmiClient(port=3000)
build = Build()
build.name = "build-1"

ci = Ci()
ci.system = 'Jenkins'
build.ci_tool = ci

target = Target()
hw = Hardware()
hw.model = "K64F"
target.hardware = hw
build.target = target
build.target.operating_system = "win32"

print(build.data)
client.post_build(build)



"""
OpenTMI python client decorators
"""
from enum import Enum
import re
import functools

def requires_logged_in(func):
    """
    Decorator which verify that client are logged in
    if not but env variables are available
    it tries to loggin using them
    :param fn: function to decorated
    :return: wrapper function
    """
    def ret_fn(*args):
        """
        wrapper function
        :param args: argument for decorated function
        :return: return decorated function return values
        """
        self = args[0]
        if not self.is_logged_in:
            self.try_login(raise_if_fail=True)
        return func(*args)
    return ret_fn

def setter_rules(value_type=str, each_type=None, enum=None, match=None):
    """
    setter rules
    :param value_type: required Type
    :param each_type: required type for each items in case of list
    :param enum: String, only allowed items
    :param match: string, regex pattern to be match
    :return: decorator function
    :raises: ValueError or TypeError
    example: require int type for setter - otherwise raise ValueError
    @count.setter
    @setter_rules(type=int)
    def count(self, value):
        self.value = value
    """
    def decorator_wrapper(func):
        """
        Decorator wrapper
        :param func: function for wrap
        :return: decorator wrapper
        """
        @functools.wraps(func)
        def function_wrapper(key, value):
            """
            Function wrapper
            :param key:
            :param value: value given for setter
            :return: target function return value
            """
            if not isinstance(value, value_type):
                raise TypeError("type of value must be an {}".format(value_type.__name__))
            if each_type:
                if not all(isinstance(i, each_type) for i in value):
                    raise TypeError("{} list values must be {}".format(key, each_type.__name__))
            elif isinstance(value, str) and match:
                if not re.match(match, value):
                    raise ValueError("Value does not match to regex: {}".format(match))
            if enum:
                members = Enum("values", enum)
                enum_values = [e.name for e in members] # pylint: disable=not-an-iterable
                if value not in enum_values:
                    raise ValueError("Value {} not in allowed list ({})".format(value, enum))
            return func(key, value)
        return function_wrapper
    return decorator_wrapper



"""
OpenTMI client exceptions
"""


class OpentmiException(Exception):
    """
    Default Opentmi Exception
    """
    def __init__(self, message):
        """
        Constructor
        :param message:
        """
        Exception.__init__(self, message)
        self.message = message


class TransportException(OpentmiException):
    """
    Transport exception
    """
    def __init__(self, message, code=None):
        """
        Constructor for Transport exceptions
        :param message: string
        :param code: status_code or None
        """
        OpentmiException.__init__(self, message)
        self.code = code



"""
OpenTMI module for File
"""
from opentmi_client.utils.Base import BaseApi
from opentmi_client.utils.decorators import setter_rules


class File(BaseApi):
    """
    File Class
    """

    def __str__(self):
        return "{}".format(self.get("name", "unknown"))

    @property
    def name(self):
        """
        Getter for file name
        :return: String
        """
        return self.get("name")

    @name.setter
    @setter_rules()
    def name(self, value):
        """
        Setter for file name
        :param value: String
        """
        self.set("name", value)

    @property
    def mime_type(self):
        """
        Getter for mime_type
        :return: String
        """
        return self.get("mime_type")

    @mime_type.setter
    @setter_rules()
    def mime_type(self, value):
        """
        Setter for mime_type
        :param value: String
        """
        self.set("mime_type", value)

    @property
    def encoding(self):
        """
        Getter for encoding
        :return: String
        """
        return self.get("encoding")

    @encoding.setter
    @setter_rules(enum='raw base64')
    def encoding(self, value):
        """
        Setter for encoding
        :param value: String
        """
        self.set("encoding", value)

    @property
    def data(self):
        """
        Getter for data
        :return: String
        """
        return self.get("data")

    @data.setter
    @setter_rules(value_type=bytearray)
    def data(self, value):
        """
        Setter for data
        :param value: bytearray
        """
        self.set("data", value)

    def set_file_data(self, value, encoding="raw"):
        """
        Set string data
        :param value: str
        :param encoding: str, "raw" by default
        """
        self.encoding = encoding
        data = bytearray()
        try:
            # python 2
            data.extend(value)
        except TypeError:
            # Python3:
            data.extend(map(ord, value))
        self.data = data



#!/usr/bin/env python
"""
Module to collect all public API's
"""
from .cli import opentmiclient_main
from .api import create, Client, OpenTmiClient, Build, Result
from .transport import Transport

if __name__ == '__main__':
    opentmiclient_main()



"""
OpenTMI SUT (Software under test) module
"""
from opentmi_client.utils.Base import BaseApi
from opentmi_client.utils.decorators import setter_rules


class Sut(BaseApi):
    """
    SUT (Software Under Test) class
    """

    def __init__(self):
        super(Sut, self).__init__()
        self.set("cut", [])
        self.set("fut", [])

    @property
    def ref(self):
        """
        Getter for reference
        :return: String
        """
        return self.get("ref")

    @ref.setter
    @setter_rules()
    def ref(self, value):
        """
        Setter for reference
        :param value: String
        """
        self.set("ref", value)

    @property
    def cut(self):
        """
        Getter for cut
        :return: String
        """
        return self.get("cut")

    @setter_rules()
    def append_cut(self, value):
        """
        Append one component under test
        :param value: String
        """
        self.cut.append(value)

    @property
    def fut(self):
        """
        Getter for fut
        :return: String
        """
        return self.get("fut")

    @setter_rules()
    def append_fut(self, value):
        """
        Append one feature under test
        :param value: String
        """
        self.get("fut").append(value)


# gitUrl: {type: String, default: ''},
# buildName: {type: String},
# buildDate: {type: Date},
# buildUrl: {type: String, default: ''},
# buildSha1: {type: String},
# branch: {type: String, default: ''},
# commitId: {type: String, default: ''},
# tag: [{type: String}],
# href: {type: String},



"""
OpenTMI module for CI details
"""
from opentmi_client.utils.Base import BaseApi
from opentmi_client.utils.decorators import setter_rules


class Location(BaseApi):
    """
    Location Class
    """
    pass


class Ci(BaseApi):
    """
    CI Class
    """
    def __init__(self, system=None, location=None):
        super(Ci, self).__init__()
        if system:
            self.system = system
        if location:
            self.location = location

    @property
    def system(self):
        """
        Getter for CI System
        :return: String
        """
        return self.get("system")

    @system.setter
    @setter_rules(value_type=str, enum="Jenkins travisCI circleCI")
    def system(self, value):
        """
        Setter for CI system
        :param value: String (Jenkins/travisCI/circleCI)
        :return: value
        """
        return self.set("system", value)

    @property
    def location(self):
        """
        Getter for CI location
        :return:
        """
        return self.get("location")

    @location.setter
    @setter_rules(value_type=Location)
    def location(self, value):
        """
        Setter for CI location
        :param value: Location
        :return: value
        """
        return self.set("location", value)



"""
Query creator module
"""
import json


class Query(object):
    """
    Base Query class
    """
    def __init__(self):
        """
        Constructor for Query
        """
        self.__query = {}

    def set(self, key, value):
        """
        Set value to query
        :param key: string
        :param value:
        :return: Query
        """
        self.__query[key] = value
        return self

    def to_string(self):
        """
        convert query to string
        :return: query as a string
        """
        return json.dumps(self.__query)



class Request(object):
    """
    Request object to create mongoose-query related url parameters
    """
    def __init__(self, request_type):
        """
        Constructor for Request
        :param request_type: type
        """
        self.__request = {
            "t": request_type,
            "q": Query()}

    @property
    def query(self):
        """
        Getter for Query object
        :return: Query
        """
        return self.__request["q"]

    def _set(self, key, value):
        self.__request[key] = value

    def _has(self, key):
        return key in self.__request

    def _push(self, key, value):
        if not self._has(key):
            self._set("f", "")
        else:
            self.__request[key] += " "
        self.__request[key] += value

    def params(self):
        """
        Get params for Transport
        :return: dict
        """
        request = self.__request.copy()
        request["q"] = request["q"].to_string()
        return request


class Find(Request):
    """
    Find Request
    """
    def __init__(self):
        """
        Constructor for Find query
        """
        Request.__init__(self, "find")

    def limit(self, limit):
        """
        Limit results items
        :param limit:
        :return: Find
        """
        self._set("l", limit)
        return self

    def skip(self, skip):
        """
        Skip results
        :param skip:
        :return: Find
        """
        self._set("sk", skip)
        return self

    def select(self, field):
        """
        Select some fields to be fetched
        :param field:
        :return: Find
        """
        self._push("f", field)
        return self



class Distinct(Request):
    """
    Distinct kind of request
    """
    def __init__(self):
        """
        Constructor for Distinct
        """
        Request.__init__(self, "distinct")

    def select(self, field):
        """
        Select field to be distinct
        :param field:
        :return: Distinct
        """
        self._set("f", field)
        return self



# pylint: disable=missing-docstring

import unittest
import os
from mock import mock, MagicMock, patch, call
from opentmi_client.api import Client, create
from opentmi_client.utils import TransportException, OpentmiException
from opentmi_client.transport.transport import Transport


def mocked_post(*args, **kwargs):
    if args[1].get("exception") == "TransportException":
        raise TransportException("")
    elif args[1].get("exception") == 'OpentmiException':
        raise OpentmiException("")
    elif args[1].get("status_code") == 404:
        return None
    return args[1]

def mocked_put(*args, **kwargs):
    if args[1].get("exception") == "TransportException":
        raise TransportException("")
    elif args[1].get("exception") == 'OpentmiException':
        raise OpentmiException("")
    elif args[1].get("status_code") == 404:
        return None
    return args[1]

def mocked_get(*args, **kwargs):
    if kwargs.get('params', {}).get('tcid') == "notfound":
        return []
    return [{"tcid": "abc"}]

def mock_transport(transport):
    transport.set_token =  MagicMock()
    transport.set_host =  MagicMock()
    transport.get_json = MagicMock()
    transport.post_json = MagicMock()
    transport.put_json = MagicMock()

DUMMY_TOKEN = "abc"


class TestClient(unittest.TestCase):

    def test_create(self):
        client = create()
        self.assertIsInstance(client, Client)

    def test_token(self):
        tr_mock = Transport()
        mock_transport(tr_mock)
        tr_mock.has_token = MagicMock()
        def has_token(*args):
            return True
        tr_mock.has_token.return_value = True
        client = Client(transport=tr_mock)
        client.set_token(DUMMY_TOKEN)
        tr_mock.set_token.assert_called_once_with(DUMMY_TOKEN)
        self.assertEqual(client.is_logged_in, True)

    def test_login(self):
        tr_mock = Transport()
        mock_transport(tr_mock)
        client = Client(transport=tr_mock)
        client.login("user", "passwd")
        tr_mock.post_json.assert_called_once_with("http://127.0.0.1/auth/login",
                                                  {"email": "user", "password": "passwd"})
    def test_login_with_access_token(self):
        tr_mock = Transport()
        mock_transport(tr_mock)
        client = Client(transport=tr_mock)
        client.login_with_access_token("token", "github")
        tr_mock.post_json.assert_called_once_with("http://127.0.0.1/auth/github/token",
                                                  {"access_token": "token"})

    def test_logout(self):
        tr_mock = Transport()
        mock_transport(tr_mock)
        client = Client(transport=tr_mock)
        client.logout()
        tr_mock.set_token.assert_called_once_with(None)

    def test_version(self):
        client = Client()
        self.assertEqual(client.get_version(), 0)

    def test_try_login_raise_when_required(self):
        client = Client()
        with self.assertRaises(OpentmiException):
            client.try_login(raise_if_fail=True)

    def test_try_login_not_raise_by_default(self):
        client = Client()
        client.try_login()

    @mock.patch.dict(os.environ, {'OPENTMI_GITHUB_ACCESS_TOKEN': 'a.b.c'})
    @patch('opentmi_client.transport.Transport.post_json', side_effect=mocked_post)
    def test_try_login_token(self,  mock_post):
        client = Client()
        mock_post.assert_called_once_with("http://127.0.0.1/auth/github/token", {"token": "a.b.c"})

    @mock.patch.dict(os.environ, {'OPENTMI_USERNAME': 'username', "OPENTMI_PASSWORD": "passw"})
    @patch('opentmi_client.transport.Transport.post_json', side_effect=mocked_post)
    def test_try_login_token(self, mock_post):
        client = Client()
        mock_post.assert_called_once_with("http://127.0.0.1/auth/login", {"email": "username", "password": "passw"})

    @patch('opentmi_client.transport.Transport.post_json', side_effect=mocked_post)
    def test_upload_build(self, mock_post):
        client = Client()
        client.set_token(DUMMY_TOKEN)
        self.assertDictEqual(client.upload_build({}), {})
        mock_post.assert_called_once_with("http://127.0.0.1/api/v0/duts/builds", {})

    @patch('opentmi_client.transport.Transport.post_json', side_effect=mocked_post)
    def test_upload_build_exceptions(self, mock_post):
        client = Client()
        client.set_token(DUMMY_TOKEN)
        self.assertEqual(client.upload_build({"exception": "TransportException"}), None)
        self.assertEqual(client.upload_build({"exception": "OpentmiException"}), None)

    @patch('opentmi_client.transport.Transport.get_json', side_effect=mocked_get)
    @patch.dict(os.environ, {'OPENTMI_GITHUB_ACCESS_TOKEN': DUMMY_TOKEN})
    @patch('opentmi_client.transport.Transport.post_json', side_effect=mocked_post)
    def test_upload_results_new_test(self, mock_post, mock_get):
        client = Client()
        tc_data = {"tcid": "notfound"}
        client.upload_results(tc_data)
        mock_get.assert_called_once_with("http://127.0.0.1/api/v0/testcases", params={"tcid": "notfound"})
        mock_post.assert_has_calls([
            #call("http://127.0.0.1/auth/github/token", {"access_token": DUMMY_TOKEN}),
            call("http://127.0.0.1/api/v0/testcases", tc_data),
            call("http://127.0.0.1/api/v0/results", tc_data, files=None)])

    @patch('opentmi_client.transport.Transport.get_json', side_effect=mocked_get)
    @patch.dict(os.environ, {'OPENTMI_GITHUB_ACCESS_TOKEN': DUMMY_TOKEN})
    @patch('opentmi_client.transport.Transport.post_json', side_effect=mocked_post)
    def test_upload_results_update_test(self, mock_post, mock_get):
        client = Client()
        tc_data = {"tcid": "abc", "hauki": {"on": "kala"}}
        client.upload_results(tc_data)
        mock_get.assert_called_once_with("http://127.0.0.1/api/v0/testcases", params={"tcid": "abc"})
        mock_post.assert_has_calls([
            # call("http://127.0.0.1/auth/github/token", {"access_token": DUMMY_TOKEN}),
            call("http://127.0.0.1/api/v0/results", tc_data, files=None)
        ])

    @patch('opentmi_client.transport.Transport.get_json', side_effect=mocked_get)
    @patch.dict(os.environ, {'OPENTMI_GITHUB_ACCESS_TOKEN': DUMMY_TOKEN})
    @patch('opentmi_client.transport.Transport.post_json', side_effect=mocked_post)
    def test_get_test(self, mock_post, mock_get):
        client = Client()
        tc_data = {"tcid": "abc %s"}
        client.get_testcases(tc_data)
        mock_get.assert_called_once_with("http://127.0.0.1/api/v0/testcases", params={"tcid": "abc %s"})
        # mock_post.assert_called_once_with("http://127.0.0.1/auth/github/token", {"access_token": DUMMY_TOKEN})



"""
Logger module for OpenTMI client
"""
import logging


def get_logger(name="OpenTMI", level=logging.DEBUG):
    """
    Get opentmi logger
    :return: Logger
    """
    # use default logger
    logger = logging.getLogger(name)
    handlers_len = len(logger.handlers)
    if handlers_len > 0:
        return logger

    if level is None:
        handler = logging.NullHandler()
    else:
        logger.setLevel(level)
        # create console handler and set level to debug
        handler = logging.StreamHandler()
        handler.setLevel(level)

        # create formatter
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')

        # add formatter to ch
        handler.setFormatter(formatter)

    # add ch to logger
    logger.addHandler(handler)
    return logger



import os
import unittest
from opentmi_client.api import Result, Execution, File


class TestResult(unittest.TestCase):
    def test_construct(self):
        result = Result()
        result.tcid = 'test'
        self.assertIsInstance(result, Result)
        self.assertEqual(result.tcid, 'test')
        self.assertEqual(result.data, {'tcid': 'test'})

    def test_verdict(self):
        result = Result()
        result.verdict = 'pass'
        self.assertEqual(result.verdict, 'pass')

    def test_execution(self):
        result = Result()
        result.execution.duration = 12.0
        self.assertEqual(result.execution.duration, 12.0)
        result.execution = Execution()
        result.execution.note = 'notes'
        self.assertEqual(result.execution.note, 'notes')
        self.assertEqual(result.execution.duration, None)

    def test_file(self):
        result = Result()
        log_file = File()
        log_file.name = "stderr.log"
        self.assertEqual(log_file.name, "stderr.log")
        log_file.set_file_data("test")
        self.assertEqual(log_file.encoding, "raw")
        self.assertEqual(str(log_file), "stderr.log")
        self.assertEqual(log_file.data.decode(), "test")
        result.execution.append_log(log_file)
        self.assertEqual(len(result.execution.logs), 1)
        self.assertEqual(result.execution.logs[0], log_file)

    def test_from_dictionary(self):
        def reducer_func(_result, value, key):
            if key == "result":
                _result.verdict = value.lower()
            elif key == "test.name":
                _result.tcid = value
            return _result
        result = Result.from_dict({"result": "PASS", "test": {"name": "hello"}}, reducer=reducer_func)
        self.assertEqual(result.data, {"exec": {"verdict": "pass"}, "tcid": "hello"})

    def test_from_junit(self):
        dir_path = os.path.dirname(os.path.realpath(__file__))
        junit_file = os.path.join(dir_path, "../test/data/junit_simple.xml")
        results = Result.from_junit_file(junit_file)
        self.assertEqual(len(results), 3)
        self.assertEqual(results[0].tcid, "should default path to an empty string")
        self.assertEqual(results[0].verdict, "fail")
        self.assertEqual(results[1].verdict, "skip")
        self.assertEqual(results[2].verdict, "pass")




"""
OpenTMI module for VCS
"""
from opentmi_client.utils.Base import BaseApi
from opentmi_client.utils.decorators import setter_rules


class Vcs(BaseApi):
    """
    Version Control System class
    """

    @property
    def name(self):
        """
        Getter for VCS Name
        :return:
        """
        return self.get("name")

    @name.setter
    @setter_rules()
    def name(self, value):
        """
        Setter for VCS name
        :param value: String
        :return: value
        """
        return self.set("name", value)

    @property
    def system(self):
        """
        Getter for VCS system, e.g. git
        :return: String
        """
        return self.get("system")

    @system.setter
    @setter_rules(enum="git SVN CSV")
    def system(self, value):
        """
        Setter for VCS system, e.g. git
        :param value: String
        :return: value
        """
        return self.set("system", value)

    @property
    def type(self):
        """
        Getter for VCS type
        :return: String
        """
        return self.get("type")

    @type.setter
    @setter_rules(enum="PR")
    def type(self, value):
        """
        Setter for VSC type, e.g. PR
        :param value: String
        :return: value
        """
        return self.set("type", value)

    @property
    def commit_id(self):
        """
        Getter for commit id
        :return: String
        """
        return self.get("commitId")

    @commit_id.setter
    @setter_rules()
    def commit_id(self, value):
        """
        Setter for commit id
        :param value: String
        :return: value
        """
        return self.set("commitId", value)

    @property
    def branch(self):
        """
        Getter for branch
        :return: String
        """
        return self.get("branch")

    @branch.setter
    @setter_rules()
    def branch(self, value):
        """
        Setter for Branch
        :param value: String
        :return: value
        """
        return self.set("branch", value)

    @property
    def base_branch(self):
        """
        Getter for base branch
        :return: String
        """
        return self.get("base_branch")

    @base_branch.setter
    @setter_rules()
    def base_branch(self, value):
        """
        Getter for base branch
        :param value: String
        :return: value
        """
        return self.set("base_branch", value)

    @property
    def base_commit(self):
        """
        Getter for base commit
        :return: String
        """
        return self.get("base_commit")

    @base_commit.setter
    @setter_rules()
    def base_commit(self, value):
        """
        Setter for base commit
        :param value: String
        :return: value
        """
        return self.set("base_commit", value)

    @property
    def pr_number(self):
        """
        Getter for PR number
        :return: String
        """
        return self.get("pr_number")

    @pr_number.setter
    @setter_rules()
    def pr_number(self, value):
        """
        Setter for PR number
        :param value: String
        :return: value
        """
        return self.set("pr_number", value)

    @property
    def url(self):
        """
        Getter for url
        :return: String
        """
        return self.get("url")

    @url.setter
    @setter_rules()
    def url(self, value):
        """
        Setter for url
        :param value: String
        :return: value
        """
        return self.set("url", value)

    @property
    def clean_wa(self):
        """
        Getter for clean workarea state
        :return: Boolean
        """
        return self.get("clean_wa")

    @clean_wa.setter
    @setter_rules(value_type=bool)
    def clean_wa(self, value):
        """
        Setter for clean workarea
        :param value: Boolean
        :return: value
        """
        return self.set("clean_wa", value)



# pylint: disable=missing-docstring

import unittest
from mock import patch, Mock
from opentmi_client.cli.main import opentmiclient_main, OpentTMIClientCLI


FAKE_TOKEN = "a.b.c"


class TestCli(unittest.TestCase):
    @patch('sys.stdout', new_callable=Mock())
    @patch("sys.exit")
    def test_help(self, mock_exit, mock_stdout):
        fake_args = ["opentmi", "--help"]
        with patch('sys.argv', fake_args):
            opentmiclient_main()
        mock_exit.assert_called_with(0)

    @patch('sys.stdout', new_callable=Mock())
    @patch("sys.exit")
    def test_version(self, mock_exit, mock_stdout):
        fake_args = ["opentmi", "version"]
        with patch('sys.argv', fake_args):
            opentmiclient_main()
        mock_exit.assert_called_with(0)

    @patch('sys.stdout', new_callable=Mock())
    @patch("sys.exit")
    def test_version_verbose(self, mock_exit, mock_stdout):
        fake_args = ["opentmi", "-v", "version"]
        with patch('sys.argv', fake_args):
            opentmiclient_main()
        mock_exit.assert_called_with(0)

    @patch('sys.stdout', new_callable=Mock())
    @patch("sys.exit")
    def test_verbose_levels(self, mock_exit, mock_stdout):

        fake_args = ["opentmi", "-v", "version"]
        with patch('sys.argv', fake_args):
            cli = OpentTMIClientCLI()

        fake_args = ["opentmi", "-vv", "version"]
        with patch('sys.argv', fake_args):
            cli = OpentTMIClientCLI()

        fake_args = ["opentmi", "-vvv", "version"]
        with patch('sys.argv', fake_args):
            cli = OpentTMIClientCLI()

        fake_args = ["opentmi", "-vvvv", "version"]
        with patch('sys.argv', fake_args):
            cli = OpentTMIClientCLI()

    @patch('opentmi_client.transport.Transport.get_json', returns=[{"tcid": "b", "name": "c"}])
    @patch('opentmi_client.transport.Transport.post_json', returns={"token": "a.b.c"})
    @patch('sys.stdout', new_callable=Mock())
    @patch("sys.exit")
    def test_list_testcases(self, mock_exit, _mock_stdout, mock_list, mock_token):
        fake_args = ["opentmi", "--token", FAKE_TOKEN, "list", "--testcases"]
        with patch('sys.argv', fake_args):
            opentmiclient_main()
        mock_exit.assert_called_with(0)

    @patch('opentmi_client.transport.Transport.get_json', returns=[{"tcid": "b", "name": "c"}])
    @patch('opentmi_client.transport.Transport.post_json', returns={"token": "a.b.c"})
    @patch('sys.stdout', new_callable=Mock())
    @patch("sys.exit")
    def test_list_campaigns(self, mock_exit, _mock_stdout, mock_token, mock_list):
        fake_args = ["opentmi", "--token", FAKE_TOKEN, "list", "--campaigns"]
        with patch('sys.argv', fake_args):
            opentmiclient_main()
        mock_exit.assert_called_with(0)

    @patch('opentmi_client.transport.Transport.get_json', returns=[{"tcid": "b", "name": "c"}])
    @patch('opentmi_client.transport.Transport.post_json', returns={"token": "a.b.c"})
    @patch('sys.stdout', new_callable=Mock())
    @patch("sys.exit")
    def test_list_results_json(self, mock_exit, _mock_stdout, mock_token, mock_list):
        fake_args = ["opentmi", "--token", FAKE_TOKEN, "list", "--testcases", "--json"]
        with patch('sys.argv', fake_args):
            opentmiclient_main()
        mock_exit.assert_called_with(0)

    @patch('opentmi_client.transport.Transport.get_json', returns=[{"tcid": "b", "name": "c"}])
    @patch('opentmi_client.transport.Transport.post_json', returns={"token": "a.b.c"})
    @patch('sys.stdout', new_callable=Mock())
    @patch("sys.exit")
    def test_list_results_json(self, mock_exit, _mock_stdout, mock_token, mock_list):
        fake_args = ["opentmi", "--token", FAKE_TOKEN, "list", "--campaigns", "--json"]
        with patch('sys.argv', fake_args):
            opentmiclient_main()
        mock_exit.assert_called_with(0)



"""
OpenTMI module for Test Framework
"""
from opentmi_client.utils.Base import BaseApi
from opentmi_client.utils.decorators import setter_rules

class Framework(BaseApi):
    """
    Framework class
    """
    def __init__(self, name=None, version=None):
        """
        Framework constructor
        :param name: String
        :param version: String
        """
        super(Framework, self).__init__()
        if name:
            self.name = name
        if version:
            self.ver = version

    @property
    def name(self):
        """
        Getter for framework name
        :return: String
        """
        return self.get("name")

    @name.setter
    @setter_rules()
    def name(self, value):
        """
        Setter for framework name
        :param value: String
        """
        self.set("name", value)

    @property
    def version(self):
        """
        Getter for framework version
        :return:
        """
        return self.get("ver")

    @version.setter
    @setter_rules()
    def version(self, value):
        """
        Setter for framework version
        :param value: String
        """
        self.set("ver", value)



# pylint: disable=missing-docstring

import unittest
from opentmi_client.utils import Query, Find, Distinct


class TestRequest(unittest.TestCase):

    def test_query_empty(self):
        self.assertEqual(Query().to_string(), "{}")

    def test_query_set(self):
        self.assertEqual(Query().set("a", "b").to_string(), '{"a": "b"}')

    def test_find_empty(self):
        self.assertDictEqual(Find().params(), {"t": "find", "q": "{}"})

    def test_find_limit(self):
        self.assertDictEqual(Find().limit(1).params(), {"t": "find", "l": 1, "q": "{}"})

    def test_find_skip(self):
        self.assertDictEqual(Find().skip(1).params(), {"t": "find", "sk": 1, "q": "{}"})

    def test_find_select(self):
        self.assertDictEqual(Find().select("aa").params(), {"t": "find", "f": "aa", "q": "{}"})

    def test_find_select_multi(self):
        self.assertDictEqual(Find()
                             .select("aa")
                             .select("bb")
                             .params(),
                             {"t": "find", "f": "aa bb", "q": "{}"})

    def test_find_query(self):
        find = Find()
        find.query.set("a", "b")
        self.assertDictEqual(find.params(), {"t": "find", "q": '{"a": "b"}'})

    def test_distinct(self):
        dist = Distinct().select("a")
        self.assertDictEqual(dist.params(), {"t": "distinct", 'q': '{}', "f": "a"})


if __name__ == '__main__':
    unittest.main()


"""
Result module
"""
from opentmi_client.api.result.Result import Result
from opentmi_client.api.result.Provider import Provider
from opentmi_client.api.result.File import File
from opentmi_client.api.result.Framework import Framework
from opentmi_client.api.result.Execution import Execution
from opentmi_client.api.result.Dut import Dut
from opentmi_client.api.result.Sut import Sut
from opentmi_client.api.result.Environment import Environment
from opentmi_client.api.result.Job import Job



"""
OpenTMI module for Test Build
"""
from opentmi_client.utils.Base import BaseApi
from opentmi_client.utils.decorators import setter_rules
from opentmi_client.api.build.Ci import Ci
from opentmi_client.api.build.Vcs import Vcs
from opentmi_client.api.build.Target import Target


class Build(BaseApi):
    """
    Build class
    """
    def __init__(self,
                 name=None):
        super(Build, self).__init__()
        self.target = Target()
        self.vcs = []
        self.ci_tool = Ci()
        if name:
            self.name = name

    @property
    def name(self):
        """
        Getter for build name
        :return: String
        """
        return self.get("name")

    @name.setter
    @setter_rules()
    def name(self, value):
        """
        Setter for build name
        :param value: String
        :return: value
        """
        self.set("name", value)

    @property
    def type(self):
        """
        Getter for build type
        :return: String
        """
        return self.get("type")

    @type.setter
    @setter_rules(value_type=str, enum="app lib test")
    def type(self, value):
        """
        Setter for build type
        :param value: String (app/lib/test)
        :return: value
        """
        self.set("type", value)

    @property
    def ci_tool(self):
        """
        Getter for CI tool
        :return: Ci
        """
        return self.get("ci")

    @ci_tool.setter
    @setter_rules(value_type=Ci)
    def ci_tool(self, value):
        """
        Getter for CI tool
        :param value: Ci instance
        :return: value
        """
        self.set("ci", value)

    @property
    def vcs(self):
        """
        Getter for VCS
        :return: Vcs
        """
        return self.get("vcs")

    @vcs.setter
    @setter_rules(value_type=list, each_type=Vcs)
    def vcs(self, value):
        """
        Setter for Vcs
        :param value: List<Vcs>
        :return: value
        """
        self.set("vcs", value)

    @property
    def uuid(self):
        """
        Getter for UUID
        :return: String
        """
        return self.get("uuid")

    @uuid.setter
    @setter_rules()
    def uuid(self, value):
        """
        Setter for uuid
        :param value: String
        :return: value
        """
        self.set("uuid", value)

    @property
    def compiled_by(self):
        """
        Getter for compiler-by
        :return: String
        """
        return self.get("compiledBy")

    @compiled_by.setter
    @setter_rules(enum="CI manual")
    def compiled_by(self, value):
        """
        Setter for compiler-by
        :param value: String (CI/manual)
        :return: value
        """
        self.set("compiledBy", value)

    @property
    def change_id(self):
        """
        Getter for change-id
        :return: String
        """
        return self.get("changeId")

    @change_id.setter
    @setter_rules()
    def change_id(self, value):
        """
        Setter for change-id
        :param value: String
        :return: value
        """
        self.set("changeId", value)

    @property
    def target(self):
        """
        Getter for target
        :return: Target
        """
        return self.get("target")

    @target.setter
    @setter_rules(value_type=Target)
    def target(self, value):
        """
        Getter for target
        :param value: Target instance
        :return: value
        """
        self.set("target", value)



"""
OpenTMI module for Result Execution
"""
from opentmi_client.utils.Base import BaseApi
from opentmi_client.utils.decorators import setter_rules
from opentmi_client.api.result.File import File
from opentmi_client.api.result.Environment import Environment
from opentmi_client.api.result.Sut import Sut


class Execution(BaseApi):
    """
    Execution class,
    holds details about test execution phase
    """
    def __init__(self,
                 verdict=None,
                 note=None,
                 duration=None,
                 environment=None):
        """
        Execution constructor
        :param verdict: String
        :param note: : String
        :param duration: float
        :param environment: Environment
        """
        super(Execution, self).__init__()
        if verdict:
            self.verdict = verdict
        if note:
            self.note = note
        if duration:
            self.duration = duration
        if environment:
            self.environment = environment
        else:
            self.environment = Environment()
        self.sut = Sut()

    @property
    def verdict(self):
        """
        Getter for test verdict
        :return: String
        """
        return self.get("verdict")

    @verdict.setter
    @setter_rules(enum='pass fail inconclusive blocked error skip')
    def verdict(self, value):
        """
        Setter for test verdict
        :param value: String (allowed values: pass, fail, inconclusive, blocked, error, skip)
        """
        self.set("verdict", value)

    @property
    def note(self):
        """
        Getter for notes. Eg notes why test fails
        :return: String
        """
        return self.get("note")

    @note.setter
    @setter_rules()
    def note(self, value):
        """
        Setter for test notes
        :param value: String
        """
        self.set("note", value)

    @property
    def duration(self):
        """
        Getter for test duration
        :return: float
        """
        return self.get("duration")

    @duration.setter
    @setter_rules(value_type=float)
    def duration(self, value):
        """
        Setter for duration
        :param value: float
        """
        self.set("duration", value)

    @property
    def profiling(self):
        """
        Getter for profiling
        :return: dict
        """
        return self.get("profiling")

    @profiling.setter
    @setter_rules(value_type=dict)
    def profiling(self, value):
        """
        Setter for profiling.
        Profiling could contains eg timespans for different test phases:
        eg: {"setUp": {"duration": 10}}
        :param value: dict
        """
        self.set("profiling", value)

    @property
    def logs(self):
        """
        Getter for logs
        :return: List<File>
        """
        return self.get("logs")

    @logs.setter
    @setter_rules(value_type=list, each_type=File)
    def logs(self, value):
        """
        Setter for logs
        :param value: List<File>
        """
        self.set("logs", value)

    @setter_rules(value_type=File)
    def append_log(self, log_file):
        """
        Appens new file to logs array
        :param log_file: File
        """
        if not isinstance(self.logs, list):
            self.logs = []
        self.logs.append(log_file)

    @property
    def environment(self):
        """
        Getter for environment
        :return: Environment
        """
        return self.get("env")

    @environment.setter
    @setter_rules(value_type=Environment)
    def environment(self, value):
        """
        Setter for environment
        :param value: Environment
        """
        self.set("env", value)

    @property
    def sut(self):
        """
        Getter for sut (Software Under Test)
        :return: Sut
        """
        return self.get("sut")

    @sut.setter
    @setter_rules(value_type=Sut)
    def sut(self, value):
        """
        Setter for sut (Software Under Test)
        :param value: Sut
        """
        self.set("sut", value)



"""
Transport module for Opentmi python client
"""
import json
import requests
from requests import Response, RequestException
try:
    # python2
    from urllib import urlencode, quote
except ImportError:
    # python3
    from urllib.parse import urlencode, quote
from opentmi_client.utils import get_logger, resolve_host, resolve_token, TransportException

REQUEST_TIMEOUT = 30
NOT_FOUND = 404


class Transport(object):
    """
    Transport class which handle communication layer
    Mostly wrappers http rest requests to more simple APIs
    """

    __request_timeout = 10

    def __init__(self, host="127.0.0.1", port=None, token=None):
        """
        Constructor for Transport
        :param host: Hostname as a String
        :param port: optional port as a positive Integer
        :param token: optional token
        """
        self.logger = get_logger()
        self.__token = None
        self.__host = None
        if token:
            self.set_token(token)

        self.set_host(host, port)
        self.logger.info("OpenTMI host: %s", self.host)

    def set_host(self, host, port=None):
        """
        Set host address and port
        :param host:
        :param port:
        :return: None
        """
        self.__host = resolve_host(host, port)
        token = resolve_token(host)
        if token:
            self.__host = self.__host.replace(token + "@", "")
            self.set_token(token)

    @property
    def host(self):
        """
        Getter for host
        :return: host as a string
        """
        return self.__host

    @property
    def token(self):
        """
        Getter for token
        :return: token as a string or None
        """
        return self.__token

    def set_token(self, token):
        """
        Set authentication token
        :param token:
        :return: Transport
        """
        self.__token = token
        return self

    def has_token(self):
        """
        Check if token is available
        :return: Boolean True if token exists, otherwise False
        """
        return self.__token is not None

    def get_url(self, path):
        """
        Create url from path
        :param path: string, e.g. "/auth/login"
        :return: url as a string
        """
        return self.__host + path

    @property
    def __headers(self):
        headers = {
            "content-type": "application/json",
            "Connection": "close"
        }
        if self.__token:
            headers["Authorization"] = "Bearer " + self.__token
        return headers

    @staticmethod
    def _params_encode(params):
        """
        Encode parameters
        :param params: Dict of url parameters
        """
        if not params:
            return params
        params = {k: quote(v) for k, v in params.items()}
        return params

    def get_json(self, url, params=None):
        """
        GET request
        :param url: url as a string
        :param params: url parameters as dict
        :raise TransportException: when something goes wrong
        :return: dict object or None if not found
        """
        try:
            self.logger.debug("GET: %s?%s", url, urlencode(params) if params else '')
            response = requests.get(url,
                                    headers=self.__headers,
                                    timeout=REQUEST_TIMEOUT,
                                    params=Transport._params_encode(params))
            if Transport.is_success(response):
                return response.json()
            if response.status_code == NOT_FOUND:
                self.logger.warning("not found")
            else:
                self.logger.warning("Request failed: %s (code: %s)",
                                    response.text, str(response.status_code))
                raise TransportException(response.text, response.status_code)
        except RequestException as error:
            self.logger.warning("Connection error %s", error)
            raise TransportException(str(error))
        except (ValueError, TypeError) as error:
            raise TransportException(str(error))
        return None

    def post_json(self, url, payload, files=None):
        """
        POST request
        :param url:
        :param payload:
        :param files:
        :return: response as dict
        """
        try:
            response = requests.post(url,
                                     json=payload,
                                     headers=self.__headers,
                                     files=files if not None else [],
                                     timeout=REQUEST_TIMEOUT)
            if Transport.is_success(response):
                data = json.loads(response.text)
                return data
            self.logger.warning("status_code: %s", str(response.status_code))
            self.logger.warning(response.text)
            raise TransportException(response.text, response.status_code)
        except RequestException as error:
            self.logger.warning(error)
            raise TransportException(str(error))
        except (ValueError, TypeError, KeyError) as error:
            raise TransportException(error)

    def put_json(self, url, payload):
        """
        PUT requests
        :param url:
        :param payload: dict
        :return: response as a dict
        """
        try:
            response = requests.put(url,
                                    json=payload,
                                    headers=self.__headers,
                                    timeout=REQUEST_TIMEOUT)
            if Transport.is_success(response):
                data = json.loads(response.text)
                return data
            self.logger.warning("status_code: %s", str(response.status_code))
            self.logger.warning(response.text)
            raise TransportException(response.text, response.status_code)
        except RequestException as error:
            self.logger.warning(error)
            raise TransportException(str(error))
        except (ValueError, TypeError) as error:
            raise TransportException(error)
        except Exception as error:
            self.logger.warning(error)
            raise TransportException(str(error))

    @staticmethod
    def is_success(response):
        """
        Check if status_code is success range
        :param response:
        :return:
        """
        assert isinstance(response, Response)
        code = response.status_code
        return 300 > code >= 200



# pylint: disable=missing-docstring

import unittest
from logging import Logger, DEBUG
from opentmi_client import OpenTmiClient
from opentmi_client.utils import get_logger

class TestLogger(unittest.TestCase):
    def test_default_logger(self):
        logger = get_logger()
        self.assertIsInstance(logger, Logger)
        self.assertEqual(len(logger.handlers), 1)

    def test_null_logger(self):
        logger = get_logger(name="new", level=None)
        self.assertIsInstance(logger, Logger)
        self.assertEqual(len(logger.handlers), 1)

    def test_set_logger(self):
        client = OpenTmiClient()
        client.set_logger(get_logger(level=DEBUG))


if __name__ == '__main__':
    unittest.main()



import unittest
from opentmi_client.utils.Base import BaseApi


class TestBase(unittest.TestCase):
    def test_basics(self):
        base = BaseApi()
        self.assertEqual(base.data, {})
        self.assertEqual(str(base), '{}')
        self.assertTrue(base.is_empty)

    def test_id(self):
        base = BaseApi()
        base._id = "abc"
        self.assertEqual(base.data, {"_id": "abc"})
        self.assertEqual(str(base), '{\n  "_id": "abc"\n}')
        self.assertFalse(base.is_empty)

    def test_set(self):
        base = BaseApi()
        base.set('a', 1)
        self.assertEqual(base.data, {'a': 1})
        self.assertEqual(str(base), '{\n  "a": 1\n}')
        self.assertFalse(base.is_empty)
        self.assertEqual(base.get('a'), 1)

    def test_unset(self):
        base = BaseApi()
        base.set('a', 1)
        base.unset('a')
        self.assertTrue(base.is_empty)

    def test_nested(self):
        base = BaseApi()
        base.set('a.b.c', 2)
        self.assertEqual(base.data, {'a': {'b': {'c': 2}}})
        self.assertEqual(str(base), '{\n  "a": {\n    "b": {\n      "c": 2\n    }\n  }\n}')
        self.assertEqual(base.get('a.b.c'), 2)

    def test_subclass(self):
        base = BaseApi()
        class Test(BaseApi):
            pass
        base.set('a.b', Test())
        base.set('b', 'b')
        self.assertEqual(base.data, {'b': 'b'})
        self.assertEqual(str(base), '{\n  "b": "b"\n}')
        self.assertFalse(base.is_empty)

    def test_set_data_key_not_exists(self):
        base = BaseApi()
        with self.assertRaises(KeyError):
            base.data = {"a": "a"}

    def test_set_data_key_exists(self):
        class A(BaseApi):
            @property
            def a(self):
                return self.get("a")
            @a.setter
            def a(self, value):
                self.set("a", value)

        base = A()
        base.data = {"a": "a"}
        self.assertEqual(base.data, {"a": "a"})
        self.assertEqual(base.a, "a")

    def test_set_data_nested_key_not_exists(self):
        class A(BaseApi):
            @property
            def a(self):
                return self.get("a")
            @a.setter
            def a(self, value):
                self.set("a", value)

        base = A()
        with self.assertRaises(KeyError):
            base.data = {"a": {"a": 1}}

    def test_set_data_nested_key_exists(self):
        class InnerA(BaseApi):
            @property
            def a(self):
                return self.get("a")
            @a.setter
            def a(self, value):
                self.set("a", value)

        class A(BaseApi):
            def __init__(self):
                super(A, self).__init__()
                self.a = InnerA()
            @property
            def a(self):
                return self.get("a")
            @a.setter
            def a(self, value):
                self.set("a", value)

        base = A()
        base.data = {"a": {"a": 1}}
        self.assertEqual(base.data, {"a": {"a": 1}})
        self.assertEqual(base.a, {"a": 1})

    def test_set_data_empty(self):
        base = BaseApi()
        base.data = {}
        self.assertEqual(base.data, {})



"""
OpenTMI Result module
"""
# standard imports
import json
# 3rd party imports
from junitparser import JUnitXml
from junitparser import Failure as JunitFailure
from junitparser import Skipped as JunitSkipped
from pydash import reduce_, map_keys
# Internal imports
from opentmi_client.utils.Base import BaseApi
from opentmi_client.utils.decorators import setter_rules
from opentmi_client.api.result.Job import Job
from opentmi_client.api.result.Execution import Execution
from opentmi_client.api.result.File import File


class Result(BaseApi):
    """
    Result class
    """
    def __init__(self,
                 tcid=None,
                 tc_ref=None):
        """
        Constructor for Result
        :param tcid: String
        :param tc_ref: String
        """
        super(Result, self).__init__()
        self.job = Job()
        self.execution = Execution()
        if tcid:
            self.tcid = tcid
        if tc_ref:
            self.tc_ref = tc_ref

    @staticmethod
    def from_dict(dictionary, reducer=None):
        """
        Create Result from plain dictionary (JSON)
        :param dictionary: Dict
        :param reducer: optional, e.g. lambda result, value, key: result
        :return: Result
        """
        def reducer_func(_result, value, key):
            """
            Inner reducer_function
            :param _result: Result object
            :param value: value from input dictionary
            :param key: key from dictionary, nested key is format: "<key>.<inner-key>"
            :return: Result
            """
            if isinstance(value, dict):
                new_value = map_keys(value, lambda inner_value, inner_key:
                                     "{}.{}".format(key, inner_key))
                return reduce_(new_value, reducer_func, _result)
            return reducer(_result, value, key)

        result = Result()
        if reducer:
            reduce_(dictionary, reducer_func, result)
        else:
            result.data = dictionary
        return result

    @staticmethod
    def from_jsonfile(json_filename, reducer=None):
        """
        Create Result from plain dictionary (JSON)
        :param json_filename: String, filename
        :param reducer: optional reducer function, see Result.from_dict
        :return: Result
        """
        data = json.load(json_filename)
        return Result.from_dict(data, reducer=reducer)

    @staticmethod
    def from_junit_file(junit_filename, wrapper=None):
        """
        Read junit file to list of Result's
        :param junit_filename: String, filename
        :param wrapper: Optional wrapper function, which is called for each case in junit file.
        Function purpose is to fill optional properties for result.
        wrapper need to set at least `tcid` which by default is set using case.name
        e.g. lambda result, case, suite: result.
        :return: [Result]
        """
        xml = JUnitXml.fromfile(junit_filename)
        results_list = []
        for suite in xml:
            # handle suites
            for case in suite:
                # handle cases
                result = Result()
                if case.system_out:
                    log_file = File()
                    log_file.data = case.system_out
                    log_file.name = "system_out"
                    result.execution.append_log(log_file)
                if case.system_err:
                    log_file = File()
                    log_file.data = case.system_err
                    log_file.name = "system_err"
                    result.execution.append_log(log_file)
                result.execution.duration = case.time
                if isinstance(case.result, JunitFailure):
                    result.execution.verdict = 'fail'
                elif isinstance(case.result, JunitSkipped):
                    result.execution.verdict = 'skip'
                else:
                    result.execution.verdict = 'pass'
                if wrapper:
                    wrapper(result, case, suite)
                else:
                    result.tcid = case.name
                    result.execution.sut.append_cut(case.classname)
                results_list.append(result)
        return results_list

    def __str__(self):
        """
        Stringify function
        :return: String
        """
        return "{} - {}".format(self.get("tcid", "?"), self.get("exec.verdict", "?"))

    @property
    def tcid(self):
        """
        Getter for test case ID
        :return: String
        """
        return self.get("tcid")

    @tcid.setter
    @setter_rules()
    def tcid(self, value):
        """
        Setter for test case ID
        :param value: String
        """
        self.set("tcid", value)

    @property
    def verdict(self):
        """
        Getter for test verdict
        :return: String
        """
        if self.execution:
            return self.execution.verdict
        return None

    @verdict.setter
    @setter_rules()
    def verdict(self, value):
        """
        Setter for test verdict
        :param value: String
        """
        if not self.execution:
            self.execution = Execution()
        self.execution.verdict = value

    @property
    def tc_ref(self):
        """
        Getter for test reference id
        :return: String
        """
        return self.get("tcRef")

    @tc_ref.setter
    @setter_rules()
    def tc_ref(self, value):
        """
        Setter for test reference
        :param value: String
        """
        self.set("tcRef", value)

    @property
    def job(self):
        """
        Getter for job
        :return: Job
        """
        return self.get("job")

    @job.setter
    @setter_rules(value_type=Job)
    def job(self, value):
        """
        Setter for job
        :param value: Job
        """
        self.set("job", value)

    @property
    def execution(self):
        """
        Getter for execution object
        :return: Execution
        """
        return self.get("exec")

    @execution.setter
    @setter_rules(value_type=Execution)
    def execution(self, value):
        """
        Setter for Execution
        :param value: Execution
        """
        self.set("exec", value)

    @property
    def campaign(self):
        """
        Getter for Campaign
        :return: String
        """
        return self.get("campaign")

    @campaign.setter
    @setter_rules()
    def campaign(self, value):
        """
        Setter for Campaign
        :param value: String
        """
        self.set("campaign", value)

    @property
    def campaign_ref(self):
        """
        Getter for campaign reference
        :return: String
        """
        return self.get("campaignRef")

    @campaign_ref.setter
    @setter_rules()
    def campaign_ref(self, value):
        """
        Setter for campaign ref
        :param value: String
        """
        self.set("campaignRef", value)
